from enum import Enum

from pydantic import BaseModel, Field

Range = tuple[int, int]


class Assembly(Enum):
    HUMAN = "GRCh38"
    RAT = "mRatBN7.2"


class EnsemblTranscript(BaseModel):
    assembly_name: Assembly
    seq_region_name: str
    start: int
    end: int
    id: str
    version: int
    display_name: str


class TranscriptId(BaseModel):
    id: str = Field(pattern=r"^ENST\d+\.\d+$")

    model_config = {
        "json_schema_extra": {
            "examples": [
                {"id": "ENST00000296930.10"},
            ]
        }
    }



import pytest

from gtgt.range import Range, intersect, overlap, subtract

# fmt: off
range_overlap = [
    # A before B
    ((0, 3), (3, 6), False),
    # B before A
    ((3, 6), (0, 3), False),
    # A is the same as B
    ((0, 3), (0, 3), True),
    # A overlaps start B
    ((0, 3), (2, 6), True),
    # A is within B, start position is the same
    ((0, 3), (0, 6), True),
    # A is within B, end position is the same
    ((3, 6), (0, 6), True),
    # A is fully within B, end position is the same
    ((1, 3), (0, 6), True),
    # A overlaps the end of B
    ((4, 7), (0, 6), True),
    # B is within A
    ((0, 6), (1, 3), True),
    # B overlaps the start of A
    ((0, 6), (0, 3), True)
]
# fmt: on


@pytest.mark.parametrize("a, b, expected", range_overlap)
def test_range_overlap(a: Range, b: Range, expected: bool) -> None:
    assert overlap(a, b) == expected
    assert overlap(b, a) == expected


intersections = [
    # Range A, range B, intersection
    ((0, 10), (10, 20), list()),
    ((0, 10), (0, 10), [(0, 10)]),
    # Test cases where A is of size 1, and B of size 3
    # In each test case, we shift A one further to the right
    ((0, 1), (1, 4), list()),
    ((1, 2), (1, 4), [(1, 2)]),
    ((2, 3), (1, 4), [(2, 3)]),
    ((3, 4), (1, 4), [(3, 4)]),
    ((4, 5), (1, 4), list()),
    # Test cases where A and B are both of size 3
    # In each test case, we shift A one further to the right
    ((0, 3), (3, 6), list()),
    ((1, 4), (3, 6), [(3, 4)]),
    ((2, 5), (3, 6), [(3, 5)]),
    ((3, 6), (3, 6), [(3, 6)]),
    ((4, 7), (3, 6), [(4, 6)]),
    ((5, 8), (3, 6), [(5, 6)]),
    ((6, 9), (3, 6), list()),
    ((7, 10), (3, 6), list()),
]


@pytest.mark.parametrize("a, b, intersection", intersections)
def test_intersect_ranges(a: Range, b: Range, intersection: list[Range]) -> None:
    assert intersect(a, b) == intersection
    assert intersect(b, a) == intersection


# fmt: off
range_subtract = [
    # The selector is before A
    #    0 1 2 3 4 5 6 7 8 9
    # A            - - - - -
    # S  - - - - -
    # E            - - - - -
    (
        # A
        [(5, 10)],
        # Selector
        [(0, 5)],
        # Expected
        [(5, 10)]
    ),
    # The selector overlaps the start of A
    #    0 1 2 3 4 5 6 7 8 9
    # A            - - - - -
    # S        - - - -
    # E                - - -
    (
        [(5, 10)],
        [(3, 7)],
        [(7, 10)],
    ),
    # The selector is contained in A
    #    0 1 2 3 4 5 6 7 8 9
    # A            - - - - -
    # S              - - -
    # E             -      -
    (
        [(5, 10)],
        [(6, 9)],
        [(5, 6), (9, 10)],
    ),
    # The selector overlaps the end of A
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - - -
    # S        - - -
    # E  - - -
    (
        [(0, 5)],
        [(3, 6)],
        [(0, 3)],
    ),
    # The selector is larger than A
    #    0 1 2 3 4 5 6 7 8 9
    # A        - - - - -
    # S      - - - - - - -
    # E
    (
        [(3, 8)],
        [(2, 9)],
        [],
    ),
    # The selector is the same as A
    #    0 1 2 3 4 5 6 7 8 9
    # A        - - - - -
    # S        - - - - -
    # E
    (
        [(3, 8)],
        [(3, 8)],
        [],
    ),
    # The selector is after A
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - - - -
    # S              - - -
    # E  - - - - - -
    (
        [(0, 6)],
        [(6, 9)],
        [(0, 6)],
    ),
    # Both A and the selector consist of multiple ranges
    # The first region in A has partial overlap with the first two selector regions
    # The second selector region (partially) overlaps the first two regions in A
    #
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - -   - -     -
    # S  - -   - - -       - - - - - -
    # E      -       -
    (
        [(0, 4), (5, 7), (9, 10)],
        [(0, 2), (3, 6), (9, 15)],
        [(2, 3), (6, 7)],
    ),
    # A is empty
    (
        [],
        [(0,2)],
        [],
    ),
    # The selector is empty, so A is unchanged
    (
        [(0, 4)],
        [],
        [(0, 4)],
    ),
    # Both A and the selector are empty
    (
        [],
        [],
        [],
    ),
    # The selector has size zero
    #    0 1 2 3 4 5 6 7 8 9
    # A  - - - -   - -     -
    # S
    # E      -       -
    (
        [(0, 4), (5, 7), (9, 10)],
        [(1,1)],
        [(0, 4), (5, 7), (9, 10)],
    ),
    # A has size zero
    (
        [(3,3)],
        [(8,10)],
        []

    ),
    # A has size zero
    (
        [(3,3)],
        [(2,4)],
        []

    ),
    # The last range in A has size zero
    (
        [(0, 4), (5, 7), (9, 10), (15,15)],
        [(0, 2), (3, 6), (9, 15)],
        [(2, 3), (6, 7)],
    ),
    # The first range in A has size zero
    (
        [(1,1), (2, 4), (5, 7), (9, 10)],
        [(0, 2), (3, 6), (9, 15)],
        [(2, 3), (6, 7)],
    ),

]
# fmt: on


@pytest.mark.parametrize("a, b, expected", range_subtract)
def test_subtract_ranges(a: list[Range], b: list[Range], expected: list[Range]) -> None:
    assert subtract(a, b) == expected



"""
Module that contains the command line app, so we can still import __main__
without executing side effects
"""

import argparse
import dataclasses
import json
import logging
import os
import secrets

from gtgt.flask import render
from gtgt.transcript import Result, Transcript

from .mutalyzer import (
    Variant,
    generate_therapies,
    init_description,
    sequence_from_description,
)
from .variant_validator import lookup_variant


def set_logging(level: str) -> None:
    format = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    format = "%(name)s - %(levelname)s - %(message)s"
    logging.basicConfig(
        format=format,
        level=level.upper(),
    )


def main() -> None:
    parser = argparse.ArgumentParser(
        description="Description of command.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--log",
        choices=["debug", "info", "warning", "error", "critical"],
        default="info",
    )

    subparsers = parser.add_subparsers(dest="command")

    transcript_parser = subparsers.add_parser(
        "transcript", help="Transcript Information"
    )

    transcript_parser.add_argument(
        "transcript_id", type=str, help="Transcript of interest"
    )

    link_parser = subparsers.add_parser("links", help="Links to external resources")
    link_parser.add_argument("hgvs_variant", type=str, help="Variant of interest")

    api_server_parser = subparsers.add_parser(
        "api_server", help="Run the GTGT API server"
    )
    api_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )

    web_server_parser = subparsers.add_parser(
        "webserver", help="Run the GTGT web server"
    )
    web_server_parser.add_argument(
        "--host", default="localhost", help="Hostname to listen on"
    )
    web_server_parser.add_argument(
        "--debug", default=False, action="store_true", help="Run Flask in debug mode"
    )

    mutator_parser = subparsers.add_parser(
        "mutate", help="Mutate the specified transcript"
    )

    mutator_parser.add_argument("transcript_id", help="The transcript to mutate")

    analyze_parser = subparsers.add_parser(
        "analyze", help="Analyze all possible exons skips for the spcified HGVS variant"
    )
    analyze_parser.add_argument(
        "hgvs", help="HGVS description of the transcript of interest"
    )

    export_parser = subparsers.add_parser(
        "export", help="Export the specified Transcript to BED format"
    )
    export_parser.add_argument(
        "hgvs", help="HGVS description of the transcript of interest"
    )

    render_parser = subparsers.add_parser(
        "render", help="Render the HTML template (helper)"
    )

    render_parser.add_argument("--results", type=str, default="")
    render_parser.add_argument("--error", "-e", help="Error payload")
    render_parser.add_argument("--variant", type=str, default="10A>T")
    args = parser.parse_args()

    set_logging(args.log)
    logger = logging.getLogger(__name__)

    if args.command == "transcript":
        d = init_description(f"{args.transcript_id}:c.=")
        t = Transcript.from_description(d)
        print(t)
    elif args.command == "links":
        logger.debug(args)
        links = lookup_variant(args.hgvs_variant).url_dict()
        for website, url in links.items():
            print(f"{website}: {url}")
    elif args.command == "api_server":
        try:
            from .app import app, uvicorn
        except ModuleNotFoundError:
            logger.critical(
                "Missing modules, please install with 'pip install gtgt[api_server]'"
            )
            exit(-1)
        uvicorn.run(app, host=args.host)
    elif args.command == "mutate":
        desc = f"{args.transcript_id}:c.="
        d = init_description(desc)
        for therapy in generate_therapies(d):
            print(f"{therapy.name}: {therapy.hgvsc}")
    elif args.command == "analyze":
        d = init_description(args.hgvs)
        transcript = Transcript.from_description(d)
        # Convert Result objects to dict
        results = [dataclasses.asdict(x) for x in transcript.analyze(args.hgvs)]
        print(json.dumps(results, indent=True, default=vars))
    elif args.command == "webserver":
        try:
            from .flask import app as flask_app
        except ModuleNotFoundError:
            logger.critical(
                f"Missing modules, please install with 'pip install gtgt[webserver]'"
            )
            exit(-1)
        if not flask_app.config.get("SECRET_KEY"):
            flask_app.secret_key = secrets.token_hex()
        flask_app.run(args.host, debug=args.debug)
    elif args.command == "export":
        # Get the transcript
        transcript = Transcript.from_description(init_description(args.hgvs))

        # Mutate the transcript
        d = init_description(args.hgvs)
        sequence = sequence_from_description(d)
        input_variants = [
            Variant.from_model(delins, sequence=sequence)
            for delins in d.delins_model["variants"]
        ]
        transcript.mutate(d, input_variants)

        for record in transcript.records():
            print(record)
    elif args.command == "render":
        if args.results:
            with open(args.results) as fin:
                file_payload = [Result.from_dict(x) for x in json.load(fin)]
        else:
            file_payload = []

        template_file = "templates/index.html.j2"
        print(
            render(
                template_file,
                variant=args.variant,
                results=file_payload,
                error=args.error,
            )
        )
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



import pytest
from pydantic import ValidationError

from gtgt.bed import Bed
from gtgt.models import TranscriptId
from gtgt.mutalyzer import HGVS
from gtgt.transcript import Transcript

payload = dict[str, str | int]


@pytest.fixture
def ucsc() -> payload:
    return {
        "chrom": "chr1",
        "chromStart": 1000,
        "chromEnd": 2000,
        "name": "ENST00000.12",
        "score": 0,
        "strand": "-",
        "thickStart": 1000,
        "thickEnd": 2000,
        "blockCount": 2,
        "blockSizes": "200,700,",
        "chromStarts": "0,300,",
        "random_field": "some nonsense",
    }


def test_HGVS_model_valid() -> None:
    """
    GIVEN a valid HGVS description
    WHEN we make an HGVS object out of it
    THEN there should be no error
    """
    HGVS(description="NM_000094.4:c.5299G>C")


INVALID_HGVS = [
    "NM_000094.4:c.5299G>",
    "NM_000094.4>",
    "NM_000094",
]


@pytest.mark.parametrize("description", INVALID_HGVS)
def test_HGVS_model_invalid(description: str) -> None:
    """
    GIVEN an invalid HGVS description
    WHEN we make an HGVS object out of itemRgb
    THEN we should get a ValidationError
    """
    with pytest.raises(ValidationError):
        HGVS(description=description)


VALID_TRANSCRIPT_ID = [
    "ENST00000296930.10",
]


@pytest.mark.parametrize("id", VALID_TRANSCRIPT_ID)
def test_TranscriptId_valid(id: str) -> None:
    TranscriptId(id=id)


INVALID_TRANSCRIPT_ID = [
    "ENST00000296930",
    "ENST00000296930.10:c.100A>T",
]


@pytest.mark.parametrize("id", INVALID_TRANSCRIPT_ID)
def test_TranscriptId_invalid(id: str) -> None:
    with pytest.raises(ValidationError):
        TranscriptId(id=id)
